import tempfile
from io import StringIO
from pathlib import Path
from textwrap import dedent
from typing import List

import cwl_utils.parser_v1_1 as parser
from cwltool.load_tool import fetch_document, resolve_and_validate_document
from dataclasses import dataclass
from WDL import parse_document

from acclimatise.yaml import yaml


@dataclass
class HelpText:
    path: str
    cmd: List[str]
    positional: int
    named: int


all_tests = [
    HelpText(path="test_data/bedtools.txt", cmd=["bedtools"], positional=43, named=1),
    HelpText(
        path="test_data/bedtools_coverage.txt",
        cmd=["bedtools", "coverage"],
        positional=0,
        named=18,
    ),
    HelpText(path="test_data/bwa.txt", cmd=["bwa"], named=0, positional=14),
    HelpText(path="test_data/bwa_mem.txt", cmd=["bwa", "mem"], positional=3, named=32),
    HelpText(
        path="test_data/bwa_bwt2sa.txt", cmd=["bwa", "bwt2sa"], named=1, positional=2
    ),
    HelpText(
        path="test_data/bwa_bwtupdate.txt",
        cmd=["bwa", "bwtupdate"],
        named=0,
        positional=1,
    ),
    HelpText(
        path="test_data/htseq_count.txt", cmd=["htseq-count"], named=14, positional=2
    ),
    HelpText(path="test_data/pisces.txt", cmd=["pisces"], named=57, positional=0),
    HelpText(
        path="test_data/podchecker.txt", cmd=["podchecker"], named=2, positional=1
    ),
    HelpText(path="test_data/samtools.txt", cmd=["samtools"], positional=28, named=0),
]


def validate_cwl(cwl: str):
    parsed = yaml.load(cwl)
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir = Path(tmpdir)
        tmpfile = tmpdir / parsed["id"]
        tmpfile.write_text(cwl)
        loading_context, workflowobj, uri = fetch_document(str(tmpfile))
        resolve_and_validate_document(loading_context, workflowobj, uri)


def validate_wdl(wdl: str):
    return parse_document(wdl)



import itertools
import unicodedata
from typing import Iterable, List

import spacy
import wordsegment

import regex as re


def sanitize_symbols(text):
    """
    Remove any non-word symbols
    """
    return re.sub("[^[:alpha:]-_., ]", "", text)


def sanitize_token(text):
    """
    Remove any non-word symbols
    """
    return re.sub("[^[:alpha:]]", "", text)


def replace_hyphens(text):
    """
    Remove any non-word symbols
    """
    return re.sub("-", "_", text)


def should_ignore(token):
    return token.pos_ not in {"NOUN", "NUM", "PRON", "PROPN", "VERB"}


def distance_to_root(token, root):
    if token == root:
        return 0
    else:
        dist = 0
        while token != root:
            dist += 1
            token = token.head
        return dist


def remove_delims(text):
    for delims in (("\[", "\]"), ("\(", "\)"), ("{", "}"), ("'", "'"), ('"', '"')):
        text = re.sub("{}.+{}".format(*delims), "", text)

    return text


def name_to_camel(words: Iterable[str]) -> str:
    """
    Converts a list of words into camelCase
    """
    words = list(words)
    cased = [words[0].lower()] + [segment.capitalize() for segment in words[1:]]
    return "".join(cased)


def name_to_snake(words: Iterable[str]) -> str:
    """
    Converts a list of words into snake_case
    """
    return "_".join([word.lower() for word in words])


def segment_string(text: str):
    """
    Divides one larger word into segments
    :param text:
    :return:
    """

    # Load wordsegment the first time
    if len(wordsegment.WORDS) == 0:
        wordsegment.load()

    base = text.lstrip("-")

    # Replace symbols with their unicode names
    translated = re.sub(
        "[^\w\s\-_]", lambda symbol: unicodedata.name(symbol[0]).lower(), base
    )

    dash_tokens = re.split("[-_ ]", translated)
    segment_tokens = itertools.chain.from_iterable(
        [wordsegment.segment(w) for w in dash_tokens]
    )
    return [sanitize_token(tok) for tok in segment_tokens]


def generate_name(description: str) -> Iterable[str]:
    """
    Given one or more sentences, attempt to parse out a concise (2-4 word) variable name
    """
    try:
        nlp = spacy.load("en")
    except IOError:
        raise Exception(
            "Spacy model doesn't exist! Install it with `python -m spacy download en`"
        )

    # Remove all delimited text
    sanitized = sanitize_symbols(remove_delims(replace_hyphens(description)))

    # Parse the sentence
    doc = nlp(sanitized)
    sentences = list(doc.sents)
    if len(sentences) == 0:
        return []

    sentence = sentences[0]
    root = sentence.root

    # Add the second most important word after the root
    candidates = sorted(
        [
            (distance_to_root(tok, root), tok)
            for tok in sentence
            if not should_ignore(tok) and root != tok
        ]
    )
    tokens = [root] if len(candidates) == 0 else [root, candidates[0][1]]

    # Add any adjectives/adverbs that describe either word
    for tok in tokens:
        for child in tok.children:
            if child.pos_ in {"ADV", "ADJ"}:
                tokens.append(child)

    # If we still don't have 3 tokens, add the 3rd most important
    if len(tokens) < 3 and len(candidates) >= 2:
        tokens.append(candidates[1][1])

    # Now sort the tokens back into their original positions
    return [
        sanitize_token(str(tok)).lower()
        for tok in sorted(tokens, key=lambda tok: tok.i)
    ]



from io import StringIO
from os import PathLike
from pathlib import Path
from typing import Generator, List

from dataclasses import dataclass

from acclimatise.converter import WrapperGenerator
from acclimatise.model import Command
from acclimatise.yaml import yaml


@dataclass
class YmlGenerator(WrapperGenerator):
    """
    Internal YML format
    """

    @classmethod
    def format(cls) -> str:
        return "yml"

    def generate_wrapper(self, cmd: Command) -> str:
        buffer = StringIO()
        yaml.dump(cmd, buffer)
        return buffer.getvalue()

    def generate_tree(
        self, cmd: Command, out_dir: PathLike
    ) -> Generator[Path, None, None]:
        out_dir = Path(out_dir)
        for cmd in cmd.command_tree():
            path = (out_dir / cmd.as_filename).with_suffix(".yml")
            with path.open("w") as fp:
                yaml.dump(cmd, fp)
            yield path



import tempfile
from itertools import product

import pytest
from pkg_resources import resource_filename

from acclimatise import WrapperGenerator, parse_help

from .util import HelpText, all_tests, validate_cwl, validate_wdl


@pytest.mark.parametrize(
    ["test", "language_data"],
    list(product(all_tests, [["cwl", validate_cwl], ["wdl", validate_wdl]])),
)
def test_all(test: HelpText, language_data):
    """
    A comprehensive end-to-end test that tests the parser and converters, using the test data files
    """
    language, validate = language_data
    with open(resource_filename("test", test.path)) as fp:
        help_text = fp.read()

    cmd = parse_help(test.cmd, help_text)

    # Check we parsed the arguments correctly
    assert len(cmd.positional) == test.positional
    assert len(cmd.named) == test.named

    # Check that the help text is included in the command
    assert cmd.help_text == help_text

    # Check the converters work
    converter = WrapperGenerator.choose_converter(language)()
    converted = converter.generate_wrapper(cmd)
    validate(converted)
